# -*- coding: utf-8 -*-
"""django-bootstrap-pagination tags."""
from django import template
from django.conf import settings
from django.core.paginator import EmptyPage
from django.core.paginator import InvalidPage
from django.core.paginator import Paginator
from django.http import Http404

register = template.Library()

DEFAULT_PAGINATION = getattr(settings, "PAGINATION_DEFAULT_PAGINATION", 20)
DEFAULT_WINDOW = getattr(settings, "PAGINATION_DEFAULT_WINDOW", 4)
DEFAULT_ORPHANS = getattr(settings, "PAGINATION_DEFAULT_ORPHANS", 0)
INVALID_PAGE_RAISES_404 = getattr(settings, "PAGINATION_INVALID_PAGE_RAISES_404", False)
INVALID_PAGE_FIXUP = getattr(settings, "PAGINATION_INVALID_PAGE_FIXUP", False)


@register.tag(name="autopaginate")
def do_autopaginate(parser, token):
    """Split the arguments to the autopaginate tag and formats them correctly."""
    split = token.split_contents()
    as_index = None
    context_var = None
    for i, bit in enumerate(split):
        if bit == "as":
            as_index = i
            break
    if as_index is not None:
        try:
            context_var = split[as_index + 1]
        except IndexError:
            raise template.TemplateSyntaxError(
                (
                    f"Context variable assignment must take the form of {{% "
                    f"{split[0]} object.example_set.all ... as context_var_name %}}"
                )
            ) from None
        del split[as_index : as_index + 2]
    if len(split) == 2:
        return AutoPaginateNode(split[1])
    if len(split) == 3:
        return AutoPaginateNode(split[1], paginate_by=split[2], context_var=context_var)
    if len(split) == 4:
        try:
            orphans = int(split[3])
        except ValueError:
            raise template.TemplateSyntaxError(
                f"Got {split[3]}, but expected integer."
            ) from None
        return AutoPaginateNode(
            split[1], paginate_by=split[2], orphans=orphans, context_var=context_var
        )
    raise template.TemplateSyntaxError(
        f"{split[0]} tag takes one required argument and one optional argument"
    )


class AutoPaginateNode(template.Node):
    """Emit the required objects to allow for Digg-style pagination.

    First, it looks in the current context for the variable specified, and using
    that object, it emits a simple ``Paginator`` and the current page object
    into the context names ``paginator`` and ``page_obj``, respectively.

    It will then replace the variable specified with only the objects for the
    current page.

    .. note::

        It is recommended to use *{% paginate %}* after using the autopaginate
        tag.  If you choose not to use *{% paginate %}*, make sure to display the
        list of available pages, or else the application may seem to be buggy.
    """

    def __init__(
        self,
        queryset_var,
        paginate_by=DEFAULT_PAGINATION,
        orphans=DEFAULT_ORPHANS,
        context_var=None,
    ):
        """Constructor."""
        self.queryset_var = template.Variable(queryset_var)
        if isinstance(paginate_by, int):
            self.paginate_by = paginate_by
        else:
            self.paginate_by = template.Variable(paginate_by)
        self.orphans = orphans
        self.context_var = context_var

    def render(self, context):
        key = self.queryset_var.var
        value = self.queryset_var.resolve(context)
        page_obj = None

        if isinstance(self.paginate_by, int):
            paginate_by = self.paginate_by
        else:
            paginate_by = self.paginate_by.resolve(context)
        paginator = Paginator(value, paginate_by, self.orphans)

        try:
            page_obj = paginator.page(context["request"].page)
        except InvalidPage as e:
            if INVALID_PAGE_RAISES_404:
                raise Http404(
                    "Invalid page requested.  If DEBUG were set to "
                    + "False, an HTTP 404 page would have been shown instead."
                ) from None

            if INVALID_PAGE_FIXUP:
                # if we're fixing up an invalid page in the request then we
                # assume that we'll goto the first page
                default_page = 1

                if type(e) == EmptyPage:
                    # page requested is out of range, as long as the page
                    # requested is greater than 1 then we show the last page
                    # otherwise just use the default (first page)
                    if int(context["request"].page) > 1:
                        # deliver the last page
                        default_page = paginator.num_pages

                # get our page_obj
                page_obj = paginator.page(default_page)

        # if we don't have a page_obj from the Paginator then bail out
        if page_obj is None:
            context[key] = []
            context["invalid_page"] = True
            return ""

        if self.context_var is not None:
            context[self.context_var] = page_obj.object_list
        else:
            context[key] = page_obj.object_list

        context["paginator"] = paginator
        context["page_obj"] = page_obj
        return ""


def paginate(context, window=DEFAULT_WINDOW, hashtag=""):
    """Render the pagination.html template.

    Args:
        context:
            Dictionary-like data structure and must contain the following keys
            paginator - a ``Paginator`` or ``QuerySetPaginator`` object.
            page_obj - the result of calling the page method on the
            aforementioned ``Paginator`` or ``QuerySetPaginator`` object, given
            the current page.
            getvars (optional) - a dictionary of all of the **GET** parameters in the current request.
            This is useful to maintain certain types of state, even when requesting
            a different page.
        window:
            Optional. Defaults to DEFAULT_WINDOW.
        hashtag:
            Optional. Defaults to "".

    Returns:
        A Digg-like display of the available pages, given the current page.
        If there are too many pages to be displayed before and after the current page, then
        elipses will be used to indicate the undisplayed gap between page numbers.
    """
    try:
        paginator = context["paginator"]
        page_obj = context["page_obj"]
        page_range = list(paginator.page_range)
        # Calculate the record range in the current page for display.
        records = {"first": 1 + (page_obj.number - 1) * paginator.per_page}
        records["last"] = records["first"] + paginator.per_page - 1
        if records["last"] + paginator.orphans >= paginator.count:
            records["last"] = paginator.count
        # First and last are simply the first *n* pages and the last *n* pages,
        # where *n* is the current window size.
        first = set(page_range[:window])
        last = set(page_range[-window:])
        # Now we look around our current page, making sure that we don't wrap
        # around.
        current_start = page_obj.number - 1 - window
        if current_start < 0:
            current_start = 0
        current_end = page_obj.number - 1 + window
        if current_end < 0:
            current_end = 0
        current = set(page_range[current_start:current_end])
        pages = []
        # If there's no overlap between the first set of pages and the current
        # set of pages, then there's a possible need for elusion.
        if len(first.intersection(current)) == 0:
            first_list = list(first)
            first_list.sort()
            second_list = list(current)
            second_list.sort()
            pages.extend(first_list)
            diff = second_list[0] - first_list[-1]
            # If there is a gap of two, between the last page of the first
            # set and the first page of the current set, then we're missing a
            # page.
            if diff == 2:
                pages.append(second_list[0] - 1)
            # If the difference is just one, then there's nothing to be done,
            # as the pages need no elusion and are correct.
            elif diff == 1:
                pass
            # Otherwise, there's a bigger gap which needs to be signaled for
            # elusion, by pushing a None value to the page list.
            else:
                pages.append(None)
            pages.extend(second_list)
        else:
            unioned = list(first.union(current))
            unioned.sort()
            pages.extend(unioned)
        # If there's no overlap between the current set of pages and the last
        # set of pages, then there's a possible need for elusion.
        if len(current.intersection(last)) == 0:
            second_list = list(last)
            second_list.sort()
            diff = second_list[0] - pages[-1]
            # If there is a gap of two, between the last page of the current
            # set and the first page of the last set, then we're missing a
            # page.
            if diff == 2:
                pages.append(second_list[0] - 1)
            # If the difference is just one, then there's nothing to be done,
            # as the pages need no elusion and are correct.
            elif diff == 1:
                pass
            # Otherwise, there's a bigger gap which needs to be signaled for
            # elusion, by pushing a None value to the page list.
            else:
                pages.append(None)
            pages.extend(second_list)
        else:
            differenced = list(last.difference(current))
            differenced.sort()
            pages.extend(differenced)
        to_return = {
            "MEDIA_URL": settings.MEDIA_URL,
            "pages": pages,
            "records": records,
            "page_obj": page_obj,
            "paginator": paginator,
            "hashtag": hashtag,
            "is_paginated": paginator.count > paginator.per_page,
        }
        if "request" in context:
            getvars = context["request"].GET.copy()
            if "page" in getvars:
                del getvars["page"]
            if len(list(getvars.keys())) > 0:
                to_return["getvars"] = "&%s" % getvars.urlencode()
            else:
                to_return["getvars"] = ""
        return to_return
    except KeyError:
        return {}


# registers the tag paginate
register.inclusion_tag("pagination.html", takes_context=True, name="paginate")(paginate)



"""paginator classes."""
from django.core.paginator import EmptyPage
from django.core.paginator import Page
from django.core.paginator import PageNotAnInteger
from django.core.paginator import Paginator


class InfinitePaginator(Paginator):
    """Paginator designed for when it's not important to know how many total pages.

    This is useful for any object_list that has no count() method or can
    be used to improve performance for MySQL by removing counts.

    The orphans parameter has been removed for simplicity and there's a link
    template string for creating the links to the next and previous pages.
    """

    def __init__(
        self,
        object_list,
        per_page,
        allow_empty_first_page=True,
        link_template="/page/%d/",
    ):
        """Constructor."""
        orphans = 0  # no orphans
        super(InfinitePaginator, self).__init__(
            object_list, per_page, orphans, allow_empty_first_page
        )
        try:
            # no count or num pages
            del self._num_pages, self._count
        except AttributeError:
            pass
        # bonus links
        self.link_template = link_template

    def validate_number(self, number):
        """Validate the given 1-based page number."""
        try:
            number = int(number)
        except ValueError:
            raise PageNotAnInteger("That page number is not an integer") from None
        if number < 1:
            raise EmptyPage("That page number is less than 1")
        return number

    def page(self, number):
        """
        Return a Page object for the given 1-based page number.
        """
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        page_items = self.object_list[bottom:top]
        # check moved from validate_number
        if not page_items:
            if number == 1 and self.allow_empty_first_page:
                pass
            else:
                raise EmptyPage("That page contains no results")
        return InfinitePage(page_items, number, self)

    def _get_count(self):
        """Return the total number of objects, across all pages."""
        raise NotImplementedError

    count = property(_get_count)

    def _get_num_pages(self):
        """Return the total number of pages."""
        raise NotImplementedError

    num_pages = property(_get_num_pages)

    def _get_page_range(self):
        """Return a 1-based range of pages.

        It is used for iterating through within a template for loop.

        Raises:
            NotImplementedError: to be implemented.
        """
        raise NotImplementedError

    page_range = property(_get_page_range)


class InfinitePage(Page):
    def __repr__(self):
        return "<Page %s>" % self.number

    def has_next(self):
        """Check for one more item than last on this page."""
        try:
            self.paginator.object_list[self.number * self.paginator.per_page]
        except IndexError:
            return False
        return True

    def end_index(self) -> int:
        """Return the 1-based index of the last object on this page.

        This index is relative to total objects found (hits).

        Returns:
            int: end index.
        """
        return (self.number - 1) * self.paginator.per_page + len(self.object_list)

    # Bonus methods for creating links
    def next_link(self):
        if self.has_next():
            return self.paginator.link_template % (self.number + 1)
        return None

    def previous_link(self):
        if self.has_previous():
            return self.paginator.link_template % (self.number - 1)
        return None


class FinitePaginator(InfinitePaginator):
    """Paginator for cases when the list of items is already finite.

    A good example is a list generated from an API call. This is a subclass
    of InfinitePaginator because we have no idea how many items exist in the
    full collection.

    To accurately determine if the next page exists, a FinitePaginator MUST be
    created with an object_list_plus that may contain more items than the
    per_page count.  Typically, you'll have an object_list_plus with one extra
    item (if there's a next page).  You'll also need to supply the offset from
    the full collection in order to get the page start_index.

    This is a very silly class but useful if you love the Django pagination
    conventions.
    """

    def __init__(
        self,
        object_list_plus,
        per_page,
        offset=None,
        allow_empty_first_page=True,
        link_template="/page/%d/",
    ):
        """Constructor."""
        super(FinitePaginator, self).__init__(
            object_list_plus, per_page, allow_empty_first_page, link_template
        )
        self.offset = offset

    def validate_number(self, number):
        super(FinitePaginator, self).validate_number(number)
        # check for an empty list to see if the page exists
        if not self.object_list:
            if number == 1 and self.allow_empty_first_page:
                pass
            else:
                raise EmptyPage("That page contains no results")
        return number

    def page(self, number):
        """Return a Page object for the given 1-based page number."""
        number = self.validate_number(number)
        # remove the extra item(s) when creating the page
        page_items = self.object_list[: self.per_page]
        return FinitePage(page_items, number, self)


class FinitePage(InfinitePage):
    def has_next(self):
        """Check for one more item than last on this page."""
        try:
            self.paginator.object_list[self.paginator.per_page]
        except IndexError:
            return False
        return True

    def start_index(self):
        """Return the 1-based index of the first object on this page.

        This index is relative to total objects in the paginator.

        Returns:
            paginator offset.
        """
        # TODO should this holler if you haven't defined the offset?
        return self.paginator.offset



"""django-pagination-bootstrap middleware."""


def get_page(self) -> int:
    """A function which will be monkeypatched onto the request to get the current integer representing the current page."""
    try:
        return int(self.GET.get("page"))
    except (KeyError, ValueError, TypeError):
        return 1


class PaginationMiddleware:
    """Pagination Middleware class.

    Inserts a variable representing the current page onto the request object if
    it exists in either **GET** or **POST** portions of the request.
    """

    def __init__(self, get_response) -> None:
        """Constructor."""
        self.get_response = get_response

    def __call__(self, request) -> None:
        request.__class__.page = property(get_page)
        response = self.get_response(request)
        return response
